//
// Gaggle
// A sort-of P2P multiplayer game environment
// by Jarno van der Linden  jvan006@cs.auckland.ac.nz
//


#include "gaggle.h"

#include <iostream>

#ifdef _WIN32
#include <windows.h> // Sleep
#endif


using namespace Gaggle;
using namespace std;

static bool MessageSendParams(const GaggleMessage& msg, PacketPriority& priority, PacketReliability& reliability, char& orderstream)
{
	switch(msg.typeID)
	{
		case ID_JOINWORLD_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 0;
			break;
		case ID_WELCOMEWORLD_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 0;
			break;
		case ID_JOINPLAYER_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 0;
			break;
		case ID_WELCOMEPLAYER_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 0;
			break;
		case ID_PLAYERACTION_MESSAGE:
			priority = HIGH_PRIORITY;
			reliability = UNRELIABLE_SEQUENCED;
			orderstream = 1;
			break;
		case ID_OBJECTSTATE_MESSAGE:
			priority = HIGH_PRIORITY;
			reliability = UNRELIABLE_SEQUENCED;
			orderstream = 1;
			break;
		case ID_OBJECTJOINED_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 3;
			break;
		case ID_OBJECTLEFT_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 3;
			break;
		case ID_ASKTEXTURE_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 0;
			break;
		case ID_TEXTURE_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 2;
			break;
		case ID_ASKOBJECTS_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 0;
			break;
		case ID_OBJECTS_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 3;
			break;
		case ID_ASKMODEL_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 0;
			break;
		case ID_MODEL_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 2;
			break;
		case ID_TEXT_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = UNRELIABLE;
			orderstream = 4;
			break;
		case ID_WORLDINTRO_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 0;
			break;
		case ID_ASKWORLDLAYOUT_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 0;
			break;
		case ID_WORLDLAYOUT_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 2;
			break;
		case ID_PLAYERREADY_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 0;
			break;
		case ID_CHANGEWORLD_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 0;
			break;
		case ID_SCORE_MESSAGE:
			priority = LOW_PRIORITY;
			reliability = RELIABLE;
			orderstream = 4;
			break;
		default:
			std::cerr << "Unknown packet send attempt type " << (int)msg.typeID << std::endl;
			return false;
	}

	return true;
}

static void Send(RakPeerInterface *peer, const GaggleMessage& msg, NetworkID remote)
{
	PacketPriority priority;
	PacketReliability reliability;
	char orderstream;

	if(MessageSendParams(msg, priority, reliability, orderstream))
		peer->Send((char *)&msg, msg.size, priority, reliability, orderstream, remote, false);
}

static void Send(RakClientInterface *client, const GaggleMessage& msg)
{
	PacketPriority priority;
	PacketReliability reliability;
	char orderstream;

	if(MessageSendParams(msg, priority, reliability, orderstream))
		client->Send((char *)&msg, msg.size, priority, reliability, orderstream);
}


// A basic Gaggle message, identified by some type ID
void GaggleMessage::Fill(GaggleMessage& msg, unsigned char typeID, unsigned int size)
{
	msg.typeID = typeID;
	msg.usetimestamp = ID_TIMESTAMP;
	msg.timestamp = RakNetGetTime();
	msg.size = size;
}


// A message describing an Identified entity
void IdentifiedMessage::Fill(IdentifiedMessage& msg, const Identified& identified, unsigned char typeID, unsigned int size)
{
	GaggleMessage::Fill(msg, typeID, size);
	msg.uid = identified.GetUID();
}

// A message containing info about a World
void WorldInfoMessage::Fill(WorldInfoMessage& msg, const HomeWorld& homeworld, unsigned char typeID, unsigned int size)
{
	GaggleMessage::Fill(msg, typeID, size);

	strncpy(msg.name, homeworld.GetName().c_str(), 64);
	msg.name[63] = '\0';
}

// A message from a HomeWorld introducing itself to a Player
void WorldIntroMessage::Fill(WorldIntroMessage& msg, const HomeWorld& homeworld, const Player& player)
{
	WorldInfoMessage::Fill(msg, homeworld, ID_WORLDINTRO_MESSAGE, sizeof(WorldIntroMessage));

	homeworld.GetNumSquares(msg.numsquares);
	msg.squaresize = homeworld.GetSquareSize();

	player.GetPos(msg.initpos);
	msg.initheading = player.GetHeading();
}

// A message from a HomeWorld requesting to be connected to a peer World
void JoinWorldMessage::Fill(JoinWorldMessage& msg, const HomeWorld& homeworld)
{
	WorldInfoMessage::Fill(msg, homeworld, ID_JOINWORLD_MESSAGE, sizeof(JoinWorldMessage));
}


// A message from a HomePlayer requesting to join a World
void JoinPlayerMessage::Fill(JoinPlayerMessage& msg, const HomePlayer& homeplayer)
{
	IdentifiedMessage::Fill(msg, homeplayer, ID_JOINPLAYER_MESSAGE, sizeof(JoinPlayerMessage));

	strncpy(msg.name, homeplayer.GetName().c_str(), 64);
	msg.name[63] = '\0';
}

// A message from a HomeWorld welcoming a just-connected World
// The message includes information about the HomeWorld
void WelcomeWorldMessage::Fill(WelcomeWorldMessage& msg, const World& world, const HomeWorld& homeworld)
{
	WorldInfoMessage::Fill(msg, homeworld, ID_WELCOMEWORLD_MESSAGE, sizeof(WelcomeWorldMessage));
}

// A message from a HomeWorld welcoming a just-connected Player
// The message includes information about the HomeWorld and the Player's UID
void WelcomePlayerMessage::Fill(WelcomePlayerMessage& msg, const Player& player, const HomeWorld& homeworld)
{
	WorldInfoMessage::Fill(msg, homeworld, ID_WELCOMEPLAYER_MESSAGE, sizeof(WelcomePlayerMessage));

	msg.uid = player.GetUID();
}

// A message from a Player to a World requesting some action to be
// done by the Player
void PlayerActionMessage::Fill(PlayerActionMessage& msg, const Player& player, ActionCode code, float value)
{
	IdentifiedMessage::Fill(msg, player, ID_PLAYERACTION_MESSAGE, sizeof(PlayerActionMessage));
	msg.code = code;
	msg.value = value;
}

// A message from a Player to a World or a World to a Player
// giving some text message
void TextMessage::Fill(TextMessage& msg, const string& text)
{
	GaggleMessage::Fill(msg, ID_TEXT_MESSAGE, sizeof(TextMessage));
	strncpy(msg.text, text.c_str(), 256);
	msg.text[255] = '\0';
}

// A message from a World to a Player giving some sort of
// scoring text
void ScoreMessage::Fill(ScoreMessage& msg, const string& text)
{
	GaggleMessage::Fill(msg, ID_SCORE_MESSAGE, sizeof(ScoreMessage));
	strncpy(msg.text, text.c_str(), 64);
	msg.text[63] = '\0';
}

// A message from a Player to a World asking about part of
// the World layout
void AskWorldLayoutMessage::Fill(AskWorldLayoutMessage& msg, int x0, int y0, int x1, int y1)
{
	GaggleMessage::Fill(msg, ID_ASKWORLDLAYOUT_MESSAGE, sizeof(AskWorldLayoutMessage));
	msg.box[0][0] = x0;
	msg.box[0][1] = y0;
	msg.box[1][0] = x1;
	msg.box[1][1] = y1;
}

// A message from a Player to a World saying the player is
// now ready to play
void PlayerReadyMessage::Fill(PlayerReadyMessage& msg)
{
	GaggleMessage::Fill(msg, ID_PLAYERREADY_MESSAGE, sizeof(PlayerReadyMessage));
}

// A message from a World to a Player asking it to change
// to another World
void ChangeWorldMessage::Fill(ChangeWorldMessage& msg, World& otherworld)
{
	GaggleMessage::Fill(msg, ID_CHANGEWORLD_MESSAGE, sizeof(ChangeWorldMessage));
	msg.otherworld = otherworld.GetNetworkID();
}

// A message containing part of a World layout
WorldLayoutMessage* WorldLayoutMessage::Create(unsigned int x0, unsigned int y0, unsigned int x1, unsigned int y1, WorldPoint*& worldpoints, WorldSquare*& worldsquares)
{
	if((x0 > x1) || (y0 > y1))
		return NULL;

	unsigned int numx = x1 - x0 + 1;
	unsigned int numy = y1 - y0 + 1;

	size_t worldpointssize = (numx+1) * (numy+1) * sizeof(WorldPoint);
	size_t worldsquaressize = numx * numy * sizeof(WorldSquare);
	unsigned int layoutmsgsize =  sizeof(WorldLayoutMessage) + worldpointssize + worldsquaressize;

	WorldLayoutMessage *layoutmsg = (WorldLayoutMessage *)new unsigned char[layoutmsgsize];
	GaggleMessage::Fill(*layoutmsg, ID_WORLDLAYOUT_MESSAGE, layoutmsgsize);

	layoutmsg->box[0][0] = x0;
	layoutmsg->box[0][1] = y0;
	layoutmsg->box[1][0] = x1;
	layoutmsg->box[1][1] = y1;

	worldpoints = (WorldPoint *)(((unsigned char *)layoutmsg) + sizeof(WorldLayoutMessage));
	worldsquares = (WorldSquare *)(((unsigned char *)worldpoints) + worldpointssize);
	return layoutmsg;
}

void WorldLayoutMessage::Destroy(WorldLayoutMessage* msg)
{
	delete[] msg;
}

//
// An identified entity in a World
// Something is identified by a unique ID
// The ID is only unique and valid within the World
//

// Construct an Identified entity
Identified::Identified()
	: uid(UNASSIGNED_UID)
{
}

// Construct an Identified entity from an info message
Identified::Identified(const IdentifiedMessage& msg)
	: uid(msg.uid)
{
}

// Identified entity destructor
Identified::~Identified()
{
}

// Get the unique identifier of the Identified entity
UID Identified::GetUID(void) const
{
	return uid;
}

// Set the unique identifier of the Identified entity
void Identified::SetUID(UID uid_in)
{
	uid = uid_in;
}


//
// A World
//

// Construct a World
World::World(const string& name_in, unsigned int numsquares_x_in, unsigned int numsquares_z_in, float squaresize_in)
	: name(name_in)
{
	numsquares[0] = numsquares_x_in;
	numsquares[1] = numsquares_z_in;
	squaresize = squaresize_in;
}

// Construct a World from a World Info message
World::World(const WorldInfoMessage& msg, NetworkID networkid_in)
	: networkid(networkid_in), name(msg.name)
{
	numsquares[0] = 0;
	numsquares[1] = 0;
	squaresize = 0;
}

// Construct a World from a World Intro message
World::World(const WorldIntroMessage& msg, NetworkID networkid_in)
	: networkid(networkid_in), name(msg.name)
{
	numsquares[0] = msg.numsquares[0];
	numsquares[1] = msg.numsquares[1];
	squaresize = msg.squaresize;
}

// World destructor
World::~World()
{
}

// Get the network ID of the World
NetworkID World::GetNetworkID(void) const
{
	return networkid;
}


const string& World::GetName(void) const
{
	return name;
}


void World::GetNumSquares(unsigned int numsquares_out[2]) const
{
	numsquares_out[0] = numsquares[0];
	numsquares_out[1] = numsquares[1];
}


float World::GetSquareSize(void) const
{
	return squaresize;
}


//
// A HomeWorld
// A HomeWorld is a World with a network peer capable of
// connecting to other Worlds
//

// Construct a HomeWorld with a given name
// The maximum number of connections includes connections to peer Worlds,
// and Players
// The network port number of the HomeWorld should usually be GAGGLE_PORT
HomeWorld::HomeWorld(const string& name, unsigned int numsquares_x, unsigned int numsquares_z, float squaresize, unsigned short maxconnects_in, unsigned short homeport_in)
	: Multiplayer<RakPeerInterface>(), World(name, numsquares_x, numsquares_z, squaresize), maxconnects(maxconnects_in), homeport(homeport_in)
{
	rakPeer = RakNetworkFactory::GetRakPeerInterface();

	rakPeer->Initialize(maxconnects, homeport, 0);
	rakPeer->SetMaximumIncomingConnections(maxconnects);
}

// HomeWorld destructor
HomeWorld::~HomeWorld()
{
	Leave();
	RakNetworkFactory::DestroyRakPeerInterface(rakPeer);
}


// Join the HomeWorld to a peer World
// The network port should usually be GAGGLE_PORT
// Returns true if the connection attempt is successful
// The connection will not actually be up until the connection
// has been accepted, which in turn will trigger an exchange
// of world info between the Worlds
bool HomeWorld::Join(const string& peerworldaddress, unsigned short peerport, unsigned long wait)
{
	bool connected = false;

	if(!peerworldaddress.empty())
	{
		connected = rakPeer->Connect((char *)peerworldaddress.c_str(), peerport, NULL, 0);
		if(connected)
		{
#ifdef _WIN32
			Sleep(wait);
#else
			usleep(wait * 1000);
#endif
		}
	}

	return connected;
}

// Disconnect the HomeWorld from all
// The HomeWorld is disconnected from all peer Worlds,
// and Players
void HomeWorld::Leave(void)
{
	rakPeer->Disconnect(300);
}


// Process any messages
void HomeWorld::ProcessPackets(void)
{
	Multiplayer<RakPeerInterface>::ProcessPackets(rakPeer);
}


// Send a message to all connected players, possibly excluding one
void HomeWorld::SendToPlayers(const GaggleMessage& msg, NetworkID exclude)
{
	for(NetworkIDPlayerMap::const_iterator pi = players.begin(); pi != players.end(); pi++)
	{
		if(pi->first != exclude)
			Send(rakPeer, msg, pi->first);
	}
}

// Send a message to a player
void HomeWorld::SendToPlayer(const GaggleMessage& msg, const Player& player)
{
	Send(rakPeer, msg, player.GetNetworkID());
}


// Process a message accepting a HomeWorld connection request to a World
// On acceptance the HomeWorld sends a message to join the World
void HomeWorld::ReceiveConnectionRequestAccepted(Packet *packet, RakPeerInterface *interfaceType)
{
	JoinWorldMessage msg;
	msg.Fill(msg, *this);
	Send(rakPeer, msg, packet->playerId);
}

// Process a message requesting a connection to the HomeWorld
// Could be from a Player or World
void HomeWorld::ReceiveNewIncomingConnection(Packet *packet, RakPeerInterface *interfaceType)
{
}

// Process a message notifying a disconnection from the HomeWorld
// The disconnection can come from a peer World, or Player
void HomeWorld::ReceiveDisconnectionNotification(Packet *packet, RakPeerInterface *interfaceType)
{
	// Who has disconnected?
	// A peer world?
	NetworkIDWorldMap::iterator pwi = peers.find(packet->playerId);
	if(pwi != peers.end())
	{
		LeftPeerWorld(pwi->second);
		peers.erase(pwi);
	}
	// A player?
	NetworkIDPlayerMap::iterator pi = players.find(packet->playerId);
	if(pi != players.end())
	{
		// Tell everyone the object has left
		ObjectLeftMessage objectleftmsg;
		objectleftmsg.Fill(objectleftmsg, *(pi->second.player));
		SendToPlayers(objectleftmsg, packet->playerId);

		PlayerHasLeft(pi->second.player);
		players.erase(pi);
	}
}

// Process a message notifying of a connection being lost
// Treat it like a normal disconnection
void HomeWorld::ReceiveConnectionLost(Packet *packet, RakPeerInterface *interfaceType)
{
	ReceiveDisconnectionNotification(packet, interfaceType);
}


// Check to see if the player can be welcomed
void HomeWorld::CheckPlayerWelcome(PlayerData& playerdata)
{
	// If player hasn't been welcomed yet, and we have model and texture,
	// and the player is ready to play, then welcome the player and tell
	// everyone about the new player
	if((!playerdata.welcomed) && playerdata.isready && playerdata.havemodel && playerdata.havetexture)
	{
		EnterPlayer(*(playerdata.player));
		WelcomePlayerMessage welcomemsg;
		welcomemsg.Fill(welcomemsg, *(playerdata.player), *this);
		SendToPlayer(welcomemsg, *(playerdata.player));
		playerdata.welcomed = true;
		ObjectJoinedMessage objectjoinedmsg;
		objectjoinedmsg.Fill(objectjoinedmsg, *(playerdata.player));
		SendToPlayers(objectjoinedmsg, playerdata.player->GetNetworkID());
	}
}

// Process any Gaggle messages
void HomeWorld::ProcessUnhandledPacket(Packet *packet, unsigned char packetIdentifier, RakPeerInterface *interfaceType)
{
	if(packetIdentifier == ID_JOINWORLD_MESSAGE)
	{
		// A World has connected and now wants to join
		// Remember the World and send back a welcome message
		const JoinWorldMessage* msg = (JoinWorldMessage *)(packet->data);
		World *world = JoinPeerWorld(*msg, packet->playerId);
		if(world)
		{
			peers.insert(NetworkIDWorldPair(packet->playerId, world));
			WelcomeWorldMessage welcomemsg;
			welcomemsg.Fill(welcomemsg, *world, *this);
			Send(rakPeer, welcomemsg, packet->playerId);
		}
	}
	else if(packetIdentifier == ID_JOINPLAYER_MESSAGE)
	{
		// A Player has connected and now wants to play with us
		// Remember the Player, and ask for model and texture
		const JoinPlayerMessage* msg = (JoinPlayerMessage *)(packet->data);
		Player *player = JoinPlayer(*msg, packet->playerId);
		if(player)
		{
			PlayerData playerdata = { player, false, false, false, false };
			players.insert(NetworkIDPlayerPair(packet->playerId, playerdata));
			// Tell a bit about me and the initial state of the Player
			WorldIntroMessage worldintromsg;
			worldintromsg.Fill(worldintromsg, *this, *player);
			SendToPlayer(worldintromsg, *player);
			// Ask for player model and texture
			AskModelMessage askmodelmsg;
			askmodelmsg.Fill(askmodelmsg, UNASSIGNED_UID);
			SendToPlayer(askmodelmsg, *player);
			AskTextureMessage asktexturemsg;
			asktexturemsg.Fill(asktexturemsg, UNASSIGNED_UID);
			SendToPlayer(asktexturemsg, *player);
		}
	}
	else if(packetIdentifier == ID_WELCOMEWORLD_MESSAGE)
	{
		// We have connected to a peer World and are now being welcomed
		// Remember the peer World
		const WelcomeWorldMessage* msg = (WelcomeWorldMessage *)(packet->data);
		World *world = JoinPeerWorld(*msg, packet->playerId);
		if(world)
		{
			peers.insert(NetworkIDWorldPair(packet->playerId, world));
		}
	}
	else if(packetIdentifier == ID_PLAYERACTION_MESSAGE)
	{
		// A Player is requesting to do some action here
		const PlayerActionMessage* msg = (PlayerActionMessage *)(packet->data);
		NetworkIDPlayerMap::iterator pi = players.find(packet->playerId);
		if(pi != players.end())
		{
			PlayerAction(*(pi->second.player), *msg);
		}
	}
	else if(packetIdentifier == ID_ASKTEXTURE_MESSAGE)
	{
		// A player is asking us about a texture
		const AskTextureMessage* msg = (AskTextureMessage *)(packet->data);
		NetworkIDPlayerMap::iterator pi = players.find(packet->playerId);
		if(pi != players.end())
		{
			const Texture* texture = AskTexture(*(pi->second.player), *msg);
			if(texture)
			{
				TextureMessage *texturemsg = TextureMessage::Create(*texture);
				SendToPlayer(*texturemsg, *(pi->second.player));
				texturemsg->Destroy(texturemsg);
			}
		}
	}
	else if(packetIdentifier == ID_TEXTURE_MESSAGE)
	{
		// A player is telling us about their texture
		const TextureMessage* msg = (TextureMessage *)(packet->data);
		NetworkIDPlayerMap::iterator pi = players.find(packet->playerId);
		if(pi != players.end())
		{
			PlayerData& playerdata = pi->second;
			PlayerTexture(*(playerdata.player), *msg);
			playerdata.havetexture = true;
			CheckPlayerWelcome(playerdata);
		}
	}
	else if(packetIdentifier == ID_ASKOBJECTS_MESSAGE)
	{
		// A player is asking us about the UIDs of all the objects in the World
		const AskObjectsMessage* msg = (AskObjectsMessage *)(packet->data);
		NetworkIDPlayerMap::iterator pi = players.find(packet->playerId);
		if(pi != players.end())
		{
			ObjectsMessage* objectsmsg = AskObjects(*(pi->second.player), *msg);
			if(objectsmsg)
			{
				SendToPlayer(*objectsmsg, *(pi->second.player));
				objectsmsg->Destroy(objectsmsg);
			}
		}
	}
	else if(packetIdentifier == ID_ASKMODEL_MESSAGE)
	{
		// A player is asking us about a model
		const AskModelMessage* msg = (AskModelMessage *)(packet->data);
		NetworkIDPlayerMap::iterator pi = players.find(packet->playerId);
		if(pi != players.end())
		{
			const Model* model = AskModel(*(pi->second.player), *msg);
			if(model)
			{
				ModelMessage *modelmsg = ModelMessage::Create(*model);
				SendToPlayer(*modelmsg, *(pi->second.player));
				modelmsg->Destroy(modelmsg);
			}
		}
	}
	else if(packetIdentifier == ID_MODEL_MESSAGE)
	{
		// A player is telling us about their model
		const ModelMessage* msg = (ModelMessage *)(packet->data);
		NetworkIDPlayerMap::iterator pi = players.find(packet->playerId);
		if(pi != players.end())
		{
			PlayerData& playerdata = pi->second;
			PlayerModel(*(playerdata.player), *msg);
			playerdata.havemodel = true;
			CheckPlayerWelcome(playerdata);
		}
	}
	else if(packetIdentifier == ID_TEXT_MESSAGE)
	{
		// A Player is sending around a text message
		const TextMessage* msg = (TextMessage *)(packet->data);
		NetworkIDPlayerMap::iterator pi = players.find(packet->playerId);
		if(pi != players.end())
		{
			PlayerTextMessage(*(pi->second.player), *msg);
		}
	}
	else if(packetIdentifier == ID_ASKWORLDLAYOUT_MESSAGE)
	{
		// A Player is asking for part of the World layout
		const AskWorldLayoutMessage* msg = (AskWorldLayoutMessage *)(packet->data);
		NetworkIDPlayerMap::iterator pi = players.find(packet->playerId);
		if(pi != players.end())
		{
			WorldLayoutMessage* layoutmsg = AskWorldLayout(*(pi->second.player), *msg);
			if(layoutmsg)
			{
				SendToPlayer(*layoutmsg, *(pi->second.player));
				layoutmsg->Destroy(layoutmsg);
			}
		}
	}
	else if(packetIdentifier == ID_PLAYERREADY_MESSAGE)
	{
		// A Player is now ready to start playing
		NetworkIDPlayerMap::iterator pi = players.find(packet->playerId);
		if(pi != players.end())
		{
			PlayerData& playerdata = pi->second;
			playerdata.isready = true;
			CheckPlayerWelcome(playerdata);
		}
	}
}


// HomeWorld has joined a peer world
// Do any needed processing
World* HomeWorld::JoinPeerWorld(const WorldInfoMessage& msg, NetworkID peer_networkid)
{
	return NULL;
}

// HomeWorld has left a peer world, either because the HomeWorld
// disconnected, or because the peer World disconnected
// Do any needed processing
void HomeWorld::LeftPeerWorld(World* world)
{
}

// HomeWorld has been joined by a Player
// Do any needed processing
Player* HomeWorld::JoinPlayer(const JoinPlayerMessage& msg, NetworkID player_networkid)
{
	return NULL;
}

// A Player that has joined has finished sending all their info
// and is about to be welcomed and introduced to the other Players
// Do any needed processing
void HomeWorld::EnterPlayer(Player& player)
{
}

// A Player has left the HomeWorld, either because the HomeWorld
// has disconnected it, or the Player has disconnected from the HomeWorld
// Do any needed processing
void HomeWorld::PlayerHasLeft(Player* player)
{
}

// A Player wants to do something in the HomeWorld
// Process the request and send an updated player state to all players
void HomeWorld::PlayerAction(Player& player, const PlayerActionMessage& msg)
{
}


// A player is telling us about their texture
// Do any needed processing
void HomeWorld::PlayerTexture(Player& player, const TextureMessage& msg)
{
}

// A player is telling us about their model
// Do any needed processing
void HomeWorld::PlayerModel(Player& player, const ModelMessage& msg)
{
}


// A player is asking us about a texture
// Do any needed processing, possibly returning a texture
const Texture* HomeWorld::AskTexture(Player& player, const AskTextureMessage& msg)
{
	return NULL;
}

// A player is asking us about the objects in the World
// Do any needed processing, possibly returning an ObjectsMessage
ObjectsMessage* HomeWorld::AskObjects(Player& player, const AskObjectsMessage& msg)
{
	return NULL;
}

// A player is asking us about a model
// Do any needed processing, possibly returning a model
const Model* HomeWorld::AskModel(Player& player, const AskModelMessage& msg)
{
	return NULL;
}

// A player has sent a us a text message
// Do any needed processing, for example broadcasting to all the players
void HomeWorld::PlayerTextMessage(Player& player, const TextMessage& msg)
{
}

// A player is asking about part of the World layout
// Do any needed processing, possibly returning a WorldLayoutMessage
WorldLayoutMessage* HomeWorld::AskWorldLayout(Player& player, const AskWorldLayoutMessage& msg)
{
	return NULL;
}



//
// A Player
//

// Construct a Player
Player::Player(const string& name_in)
	: Object(), networkid(UNASSIGNED_PLAYER_ID), name(name_in)
{
}

// Construct a Player from a Player info message
Player::Player(const JoinPlayerMessage& msg, NetworkID networkid_in)
	: Object(msg), networkid(networkid_in), name(msg.name)
{
}

// Player destructor
Player::~Player()
{
}

// Get the network ID of the Player
NetworkID Player::GetNetworkID() const
{
	return networkid;
}

// Get the name of the player
const string& Player::GetName() const
{
	return name;
}


//
// A HomePlayer
// A HomePlayer is a Player with a network client capable of
// connecting to Worlds
//

// Construct a Player with a given name
HomePlayer::HomePlayer(const string& name)
	: Multiplayer<RakClientInterface>(), Player(name) 
{
	rakClient = RakNetworkFactory::GetRakClientInterface();
}

// HomePlayer destructor
HomePlayer::~HomePlayer()
{
	Leave();
	RakNetworkFactory::DestroyRakClientInterface(rakClient);
}


// Join the Player to a World
// The network port should usually be GAGGLE_PORT
// Returns true if the connection attempt is successful
// The connection will not actually be up until the connection
// has been accepted, which in turn will trigger an exchange
// of Player and World info between the Player and World
bool HomePlayer::Join(const string& worldaddress, unsigned short port, unsigned long wait)
{
	bool connected = false;

	if(!worldaddress.empty())
	{
		connected = rakClient->Connect((char *)worldaddress.c_str(), port, 0, 0, false);
		if(connected)
		{
			connected = false;
			unsigned long endtime = RakNetGetTime() + wait;
			while((!rakClient->IsConnected()) && (RakNetGetTime() < endtime))
			{
#ifdef _WIN32
				Sleep(30);
#else
				usleep(30 * 1000);
#endif
			}
			if(rakClient->IsConnected())
				connected = true;
		}
	}

	return connected;
}

bool HomePlayer::Join(NetworkID worldnetworkid)
{
	if(worldnetworkid != UNASSIGNED_PLAYER_ID)
	{
		std::string worldaddress(rakClient->PlayerIDToDottedIP(worldnetworkid));
		unsigned short port = worldnetworkid.port;
		return Join(worldaddress, port);
	}

	return false;
}


// Disconnect the Player from a World
void HomePlayer::Leave(void)
{
	rakClient->Disconnect(300);
}


// Process any messages
void HomePlayer::ProcessPackets(void)
{
	Multiplayer<RakClientInterface>::ProcessPackets(rakClient);
}


// Send a message to the World the Player is currently in
void HomePlayer::SendToWorld(const GaggleMessage& msg)
{
	Send(rakClient, msg);
}


// Process a message accepting a request by the Player to connect to a World
// On acceptance the Player sends a message to join the World
void HomePlayer::ReceiveConnectionRequestAccepted(Packet *packet, RakClientInterface *interfaceType)
{
	networkid = rakClient->GetPlayerID();
	JoinPlayerMessage msg;
	msg.Fill(msg, *this);
	SendToWorld(msg);
}

// Process a message notifying a disconnection from the World
void HomePlayer::ReceiveDisconnectionNotification(Packet *packet, RakClientInterface *interfaceType)
{
	LeftWorld();
	networkid = UNASSIGNED_PLAYER_ID;
}

// Process any Gaggle messages
void HomePlayer::ProcessUnhandledPacket(Packet *packet, unsigned char packetIdentifier, RakClientInterface *interfaceType)
{
	if(packetIdentifier == ID_WELCOMEPLAYER_MESSAGE)
	{
		// We have connected to a World, sent our play request,
		// and are now being welcomed
		const WelcomePlayerMessage* msg = (WelcomePlayerMessage *)(packet->data);
		JoinedWorld(*msg);
	}
	else if(packetIdentifier == ID_OBJECTSTATE_MESSAGE)
	{
		// An update about the state of an Object in the same
		// World as me. Possibly that Object is me!
		ObjectStateMessage* msg = (ObjectStateMessage *)(packet->data);
		// Fix timestamps
		// lt = st - dt
		long deltatime = msg->worldtime - msg->timestamp;
		if(msg->starttime != ULONG_MAX)
			msg->starttime -= deltatime;
		if(msg->endtime != ULONG_MAX)
			msg->endtime -= deltatime;
		if(msg->animstarttime != ULONG_MAX)
			msg->animstarttime -= deltatime;

		ObjectState(*msg);
	}
	else if(packetIdentifier == ID_OBJECTJOINED_MESSAGE)
	{
		const ObjectJoinedMessage* msg = (ObjectJoinedMessage *)(packet->data);
		ObjectJoined(*msg);
	}
	else if(packetIdentifier == ID_OBJECTLEFT_MESSAGE)
	{
		const ObjectLeftMessage* msg = (ObjectLeftMessage *)(packet->data);
		ObjectLeft(*msg);
	}
	else if(packetIdentifier == ID_ASKTEXTURE_MESSAGE)
	{
		const AskTextureMessage* msg = (AskTextureMessage *)(packet->data);
		const Texture* texture = AskTexture();
		if(texture)
		{
			TextureMessage *texturemsg = TextureMessage::Create(*texture);
			SendToWorld(*texturemsg);
			texturemsg->Destroy(texturemsg);
		}
	}
	else if(packetIdentifier == ID_TEXTURE_MESSAGE)
	{
		const TextureMessage* msg = (TextureMessage *)(packet->data);
		WorldTexture(*msg);
	}
	else if(packetIdentifier == ID_OBJECTS_MESSAGE)
	{
		const ObjectsMessage* msg = (ObjectsMessage *)(packet->data);
		WorldObjects(*msg);
	}
	else if(packetIdentifier == ID_ASKMODEL_MESSAGE)
	{
		const AskModelMessage* msg = (AskModelMessage *)(packet->data);
		const Model* model = AskModel();
		if(model)
		{
			ModelMessage *modelmsg = ModelMessage::Create(*model);
			SendToWorld(*modelmsg);
			modelmsg->Destroy(modelmsg);
		}
	}
	else if(packetIdentifier == ID_MODEL_MESSAGE)
	{
		const ModelMessage* msg = (ModelMessage *)(packet->data);
		WorldModel(*msg);
	}
	else if(packetIdentifier == ID_TEXT_MESSAGE)
	{
		// The World is sending us a text message
		const TextMessage* msg = (TextMessage *)(packet->data);
		WorldTextMessage(*msg);
	}
	else if(packetIdentifier == ID_SCORE_MESSAGE)
	{
		// The World is sending us some score info
		const ScoreMessage* msg = (ScoreMessage *)(packet->data);
		WorldScoreMessage(*msg);
	}
	else if(packetIdentifier == ID_WORLDINTRO_MESSAGE)
	{
		// The World is sending us some basic info about itself
		const WorldIntroMessage* msg = (WorldIntroMessage *)(packet->data);
		WorldIntro(*msg);
	}
	else if(packetIdentifier == ID_WORLDLAYOUT_MESSAGE)
	{
		// The World is sending us some requested layout
		const WorldLayoutMessage* msg = (WorldLayoutMessage *)(packet->data);
		WorldLayout(*msg);
	}
	else if(packetIdentifier == ID_CHANGEWORLD_MESSAGE)
	{
		// The World is asking us to change to another World
		const ChangeWorldMessage* msg = (ChangeWorldMessage *)(packet->data);
		ChangeWorld(*msg);
	}
}


// Player has joined a world
// Do any needed processing
void HomePlayer::JoinedWorld(const WelcomePlayerMessage& msg)
{
}

// The Player has left the World, either because the World has disconnected,
// or the Player has disconnected from the World
// Do any needed processing
void HomePlayer::LeftWorld(void)
{
}

// The Player is informed about the objects in a World
// Do any needed processing
void HomePlayer::WorldObjects(const ObjectsMessage& msg)
{
}

// Another Object has joined the same World as the Player is in
// Do any needed processing
void HomePlayer::ObjectJoined(const ObjectJoinedMessage& msg)
{
}

// Another Object in the same World as the Player has left
// Do any needed processing
void HomePlayer::ObjectLeft(const ObjectLeftMessage& msg)
{
}

// The state of an Object is updated. That Object may be me as well
// Do any needed processing
void HomePlayer::ObjectState(const ObjectStateMessage& msg)
{
}

// The Player is asked its texture
// Do any needed processing, possibly returning a texture
const Texture* HomePlayer::AskTexture(void)
{
	return NULL;
}

// The Player is asked its model
// Do any needed processing, possibly returning a model
const Model* HomePlayer::AskModel(void)
{
	return NULL;
}

// The Player is informed about a texture
// Do any needed processing
void HomePlayer::WorldTexture(const TextureMessage& msg)
{
}

// The Player is informed about a model
// Do any needed processing
void HomePlayer::WorldModel(const ModelMessage& msg)
{
}


// The player has received a text message from the World
// Do any needed processing, such as displaying it on screen
void HomePlayer::WorldTextMessage(const TextMessage& msg)
{
}

// The player has received a text message from the World
// Do any needed processing, such as displaying it on screen
void HomePlayer::WorldScoreMessage(const ScoreMessage& msg)
{
}

// The player has received a world intro message from the World
// Do any needed processing
void HomePlayer::WorldIntro(const WorldIntroMessage& msg)
{
}


// The player has received world layout after a request
// Do any needed processing
void HomePlayer::WorldLayout(const WorldLayoutMessage& msg)
{
}


// The world is asking the player to switch to a different
// world. The new world is identified by a networkid.
void HomePlayer::ChangeWorld(const ChangeWorldMessage& msg)
{
}
