//
// Gaggle
// A multi-player cooperative gaming environment
//
// The networking is provided by RakNet  http://www.rakkarsoft.com/
// being used under the GNU General Public License
//

#include <vector>
#include <string>
#include <map>

#include "Multiplayer.h"
#include "NetworkObject.h"
#include "RakServerInterface.h"
#include "RakClientInterface.h"
#include "RakNetworkFactory.h"
#include "NetworkTypes.h"


//
// The game universe consists of a number of Worlds.
// The World owned by a user is called the Home World.
// Each World is controlled by a network server run by the owner.
// What happens in a World is controlled by the server.
// A World can contain one or more links to other peer Worlds.
//
// A World has a static environment called the World Layout.
// The World Layout is defined over a 2D regular grid. It
// consists of axis-aligned single-sided walls.
// A wall may be partially or fully open.
// Each grid square has four walls, one along each of its
// sides. Walls face inwards.
// Each grid square may have a floor and ceiling.
//
// Within a World there may be a number of Objects.
// An Object has a geometric Model and a Texture.
// Multiple Objects may share the same Model and/or Texture.
// The Models and Textures of Objects are provided by the
// World.
//
// Objects, Models, and Textures have unique IDs (UID).
// A UID is unique within the World. UIDs are only valid
// within that World.
//
// A Player is an Object whose movement is dictated by
// a (possibly remote) user.
// A Player can be in only one World, and can move between
// Worlds.
// The Player owned by a user is called the Home Player.
//
// A World has ultimate control over all movement of all
// Objects within it (including Players). Players may request
// to move, but the World is free to modify the request.
//
// Like other Objects, Players have a Model and a Texture.
// Unlike other Objects, the Model and Texture for a Player
// are provided by the Player's user to the World.
//
// To other Players, a Player appears as any other Object
// in the World.
//
// Players can only communicate with the World they are
// currently in.
//
// A World can ask a Player within it to move to a peer World.
// How a peer World appears in the World (if at all), is up
// to the World.
//
// For a Player to join a World is a multiple step process:
// 1) The Player connects to the World's server
// 2) The Player receives an acceptance from the World's server
//    The Player sends a requests to join the World
// 3) The World receives the join request
//    The World sends an intro message to the Player
//    The World sends a message to the Player asking for its
//    Model and Texture data
// 4) The Player receives the intro message
//    The Player receives Model and Texture request messages
//    The Player sends its Model and Texture to the World
//    (The Player requests World Layout data)
// 5) The World receives the Player's Model and Texture data
//    (The World receives the World Layout data request)
//    (The World sends the World Layout data to the Player)
// 6) (The Player receives the World Layout data)
//    The Player sends a ready message to the World
// 7) The World receives the ready message from the Player
//    The World sends a Welcome message to the Player
// 8) The Player receives the Welcome message
//
// The step in brackets are not crucial. The World will welcome
// a Player as soon as the World has the Player's model and texture
// and receives a Ready message from the Player.
//
// Objects can have their position and rotation fixed relative to
// the rendering camera. Obviously when position or rotation are
// fixed to the camera, the Object may appear with a different
// rotation or position for different Players.
// Fixing the rotation relative to the camera makes the Object
// appear as a billboard facing the camera's projection plane.
// Fixing the position relative to the camera makes the Object
// appear to be stuck to the Player. This can be used to make
// skyboxes for example.
//

namespace Gaggle
{

// Use the RakNet PlayerID as a network ID
typedef PlayerID NetworkID;

// The default port for HomeWorld peers
const int GAGGLE_PORT = 0x4747;

// An identifier that is unique in a World, used to identify
// objects (and players), models, and textures
typedef unsigned int UID;
const UID UNASSIGNED_UID = UINT_MAX;


// Gaggle message types
enum { ID_JOINWORLD_MESSAGE = 128,
		 ID_WELCOMEWORLD_MESSAGE,
		 ID_JOINPLAYER_MESSAGE,
		 ID_WELCOMEPLAYER_MESSAGE,
		 ID_PLAYERACTION_MESSAGE,
		 ID_OBJECTSTATE_MESSAGE,
		 ID_OBJECTJOINED_MESSAGE,
		 ID_OBJECTLEFT_MESSAGE,
		 ID_ASKTEXTURE_MESSAGE,
		 ID_TEXTURE_MESSAGE,
		 ID_ASKOBJECTS_MESSAGE,
		 ID_OBJECTS_MESSAGE,
		 ID_ASKMODEL_MESSAGE,
		 ID_MODEL_MESSAGE,
		 ID_TEXT_MESSAGE,
		 ID_WORLDINTRO_MESSAGE,
		 ID_ASKWORLDLAYOUT_MESSAGE,
		 ID_WORLDLAYOUT_MESSAGE,
		 ID_PLAYERREADY_MESSAGE,
		 ID_CHANGEWORLD_MESSAGE,
		 ID_SCORE_MESSAGE,
};

// Gaggle message
#pragma pack(push)
#pragma pack(1)
struct GaggleMessage
{
	static void Fill(GaggleMessage& msg, unsigned char typeID, unsigned int size = sizeof(GaggleMessage));

	unsigned char usetimestamp;
	unsigned long timestamp;
	unsigned char typeID;
	unsigned int size;
};
#pragma pack(pop)

// A thing uniquely identified in a World
struct IdentifiedMessage;
class Identified
{
	protected:
		Identified();
	public:
		Identified(const IdentifiedMessage& msg);
		virtual ~Identified();
		UID GetUID(void) const;
		void SetUID(UID uid);
	protected:
		UID uid;		// Unique (in the World) ID of the thing
};

// A Gaggle message about an Identified object
#pragma pack(push)
#pragma pack(1)
struct IdentifiedMessage : public GaggleMessage
{
	static void Fill(IdentifiedMessage& msg, const Identified& identified, unsigned char typeID, unsigned int size = sizeof(IdentifiedMessage));

	UID uid;
};
#pragma pack(pop)


struct ModelDataMessage;
class ModelData : public Identified
{
	public:
		ModelData();
		ModelData(const unsigned char* data, size_t datasize);
		ModelData(const std::string& filename_in);
		ModelData(const ModelDataMessage& msg);
		virtual ~ModelData();

		void AddModelData(const unsigned char* data, size_t datasize);

		size_t GetCompressedSize() const;
		const unsigned char *GetCompressedData(void) const;
	protected:
		bool LoadData(const std::string& filename);
		virtual bool DecompressFromMem(const unsigned char* data, const size_t datasize) = 0;
	protected:
		std::vector<unsigned char> compresseddata;
};

#pragma pack(push)
#pragma pack(1)
// A Gaggle message containing some model data
struct ModelDataMessage : public IdentifiedMessage
{
	static ModelDataMessage* Create(const ModelData& modeldata, unsigned char typeID, unsigned int size);
	static void Destroy(ModelDataMessage* msg);

	unsigned int compressedsize;
	// Followed by sub-class data and compressedsize bytes of compressed data
};
#pragma pack(pop)

struct TextureMessage;
class Texture : public ModelData
{
	public:
		Texture(const std::string& filename);
		Texture(const TextureMessage& msg);
		virtual ~Texture();
		void GetDimension(int& width, int& height) const;
	protected:
		virtual bool DecompressFromMem(const unsigned char* data, const size_t datasize);
	protected:
		int width, height;
		unsigned char *texture;
};

#pragma pack(push)
#pragma pack(1)
// A Gaggle message containing a texture
struct TextureMessage : public ModelDataMessage
{
	static TextureMessage* Create(const Texture& texture);
};

// A Gaggle message requesting a texture
struct AskTextureMessage : public GaggleMessage
{
	static void Fill(AskTextureMessage& msg, UID uid);

	UID uid;
};
#pragma pack(pop)

struct Vector
{
	float x,y,z;
};

struct UV
{
	float u,v;
};

struct Triangle
{
	int vertidx[3];
	int texuvidx[3];
};

struct ModelMessage;
enum {
	MODELTYPE_STATIC, MODELTYPE_MD2
};
class Model : public ModelData
{
	public:
		Model(int modeltype);
		Model(unsigned char* data, size_t datasize, int modeltype);
		Model(const std::string& filename, int modeltype);
		Model(const ModelMessage& msg);
		virtual ~Model();

		int GetModelType(void) const;

		virtual void Render(void);

	protected:
		int modeltype;
};
#pragma pack(push)
#pragma pack(1)
// A Gaggle message containing a model
struct ModelMessage : public ModelDataMessage
{
	static ModelMessage* Create(const Model& model);

	int modeltype;
	// Followed by compressedsize bytes of compressed data
};

// A Gaggle message requesting a model
struct AskModelMessage : public GaggleMessage
{
	static void Fill(AskModelMessage& msg, UID uid);

	UID uid;
};
#pragma pack(pop)

class StaticModel : public Model
{
	public:
		StaticModel(const Vector* verts, const UV* texuv, unsigned int numverts, const int* idx, unsigned int numidx, unsigned int primtype);
		StaticModel(const ModelMessage& msg);
		virtual ~StaticModel();

	protected:
		virtual bool DecompressFromMem(const unsigned char* data, const size_t datasize);

	protected:
		static unsigned int magic;

		unsigned int primtype;
		unsigned int numverts;
		unsigned int numidx;
		const Vector* verts;
		const UV* texuv;
		const int* idx;
};


class MD2Model : public Model
{
	public:
		MD2Model(const std::string& filename);
		MD2Model(const ModelMessage& msg);
		virtual ~MD2Model();

	protected:
		virtual bool DecompressFromMem(const unsigned char* data, const size_t datasize);

	protected:
		std::vector<std::vector<Vector> > verts;
		std::vector<UV> texuv;
		std::vector<Triangle> triangles;

		int numglcommands;
		std::vector<int> glcommands;
};


struct ObjectJoinedMessage;
struct ObjectStateMessage;
class Object : public Identified
{
	protected:
		Object();
	public:
		Object(const ObjectJoinedMessage& msg);
		Object(const IdentifiedMessage& msg);
		virtual ~Object();

		unsigned long GetStartTime(void) const;
		unsigned long GetEndTime(void) const;
		void GetPos(float& x, float& y, float& z) const;
		void GetPos(Vector& p) const;
		void GetDPos(float& dx, float& dy, float& dz) const;
		void GetDPos(Vector& dp) const;
		void GetDDPos(float& ddx, float& ddy, float& ddz) const;
		void GetDDPos(Vector& ddp) const;
		void GetExtrapolatedPos(unsigned long curtime, float& x, float& y, float& z) const;
		void GetExtrapolatedPos(unsigned long curtime, Vector& p) const;
		void GetExtrapolatedDPos(unsigned long curtime, float& dx, float& dy, float& dz) const;
		void GetExtrapolatedDPos(unsigned long curtime, Vector& dp) const;
		float GetHeading(void) const;
		float GetDHeading(void) const;
		float GetExtrapolatedHeading(unsigned long curtime) const;
		void GetAnimLoop(unsigned int& startframe, unsigned int& endframe) const;
		float GetAnimFPS(void) const;
		unsigned long GetAnimStartTime(void) const;
		float GetExtrapolatedFrame(unsigned long curtime);

		void SetStartTime(unsigned long starttime);
		void SetEndTime(unsigned long endtime);
		void SetPos(float x, float y, float z);
		void SetPos(const Vector& p);
		void SetDPos(float dx, float dy, float dz);
		void SetDPos(const Vector& dp);
		void SetDDPos(float ddx, float ddy, float ddz);
		void SetDDPos(const Vector& ddp);
		void SetHeading(float heading);
		void SetDHeading(float dheading);
		void SetAnimLoop(unsigned int startframe, unsigned int endframe);
		void SetAnimFPS(float animfps);
		void SetAnimStartTime(unsigned long animstarttime);

		void SetState(const ObjectStateMessage& msg);

		UID GetModelUID(void) const;
		UID GetTextureUID(void) const;
		void SetModelUID(UID uid);
		void SetTextureUID(UID uid);

		void GetCamFixed(bool& camfixedpos, bool& camfixedrot) const;
		void SetCamFixed(bool camfixedpos, bool camfixedrot);

	protected:
		UID modeluid;
		UID textureuid;

		bool camfixedpos;
		bool camfixedrot;

		// Basic object state
		unsigned long starttime;	// Time at which the state was set, in ms
		unsigned long endtime;		// Time at which movement should stop, ULONG_MAX if never, in ms
		Vector pos;						// Position of player in metres
		Vector dpos;					// Velocity in m/s
		Vector ddpos;					// Acceleration in m/(s^2)
		float heading;					// Heading the object is facing in radians, 0 is +Z direction
		float dheading;				// Change in heading in radians/s, +ve is turn left, -ve right

		unsigned int startframe;	// First frame of animation loop
		unsigned int endframe;		// Last frame of animation loop
		float animfps;					// Animation frames/second
		unsigned long animstarttime;	// Start time of animation loop
};

#pragma pack(push)
#pragma pack(1)
// A Gaggle message informing about an Object state
struct ObjectStateMessage : public IdentifiedMessage
{
	static void Fill(ObjectStateMessage& msg, const Object& object);

	unsigned long worldtime;
	unsigned long starttime;
	unsigned long endtime;
	Vector pos;
	Vector dpos;
	Vector ddpos;
	float heading;
	float dheading;

	unsigned int startframe;
	unsigned int endframe;
	float animfps;
	unsigned long animstarttime;
};

// A Gaggle message informing about a new object in the World
struct ObjectJoinedMessage : public IdentifiedMessage
{
	static void Fill(ObjectJoinedMessage&, const Object& object);
	UID textureuid;
	UID modeluid;
	unsigned char camfixed;	// Bit 0: camfixedpos, bit1: camfixedrot
};
// A Gaggle message informing about an object having left the World
struct ObjectLeftMessage : public IdentifiedMessage
{
	static void Fill(ObjectLeftMessage& msg, const Object& object);
};
// A Gaggle message asking for the UIDs of all the objects in the World
struct AskObjectsMessage : public GaggleMessage
{
	static void Fill(AskObjectsMessage& msg);
};
struct ObjectData
{
	UID uid;
	UID modeluid;
	UID textureuid;
};
struct ObjectsMessage : public GaggleMessage
{
	static ObjectsMessage* Create(unsigned int numobjects, ObjectData*& objectdata);
	static void Destroy(ObjectsMessage* msg);

	unsigned int numobjects;
	// Followed by numobjects ObjectData, one for each object
};
#pragma pack(pop)

// A World
struct WorldInfoMessage;
struct WorldIntroMessage;
class World
{
	protected:
		World(const std::string& name, unsigned int numsquares_x, unsigned int numsquares_z, float squaresize);
	public:
		World(const WorldInfoMessage& msg, NetworkID networkid);
		World(const WorldIntroMessage& msg, NetworkID networkid);
		virtual ~World();
		NetworkID GetNetworkID(void) const;
		const std::string& GetName(void) const;

		void GetNumSquares(unsigned int numsquares[2]) const;
		float GetSquareSize(void) const;

	protected:
		NetworkID networkid;
		std::string name;

		unsigned int numsquares[2];	// Number of squares in X and Z direction
		float squaresize;					// Size of a square in metres
};

#pragma pack(push)
#pragma pack(1)

/* The World layout is composed of a regular grid.
 * The grid is numsquares[0] across in the +X direction
 * and numsquares[1] deep in the +Z direction.
 * The corner of the grid is at the origin.
 * The grid has a uniform scale factor which defined the size in
 * metres of a single grid square. The grid therefore spans a
 * rectangle of (numsquares[0]*squaresize, numsquares[1]*squaresize)
 * metres. This is the range of World coordinates.
 *
 *    -------> +X
 *  (0,0)
 * |  +-----+-----+-----+
 * |  |     |     |     |
 * |  |     |     |     |
 * V  +-----+-----+-----+
 * +Z |     |     |     |
 *    |     |     |     |
 *    +-----+-----+-----+
 *                     (3*squaresize, 2*squaresize)
 *
 *
 * Each grid point defines four heights:
 *  0 : The height from 0 to the floor
 *  1 : The height from 0 to the bottom of an opening
 *  2 : The height from 0 to the top of an opening
 *  3 : The height from 0 to the ceiling
 * The heights are in increasing order: h[0] <= h[1] <= h[2] <= h[3]
 *
 * Each side of a grid square defines a wall.
 * Walls are single sided, and face inwards towards the centre
 * of the square. Walls are not shared between grid squares.
 * Each grid square defines its own set of four walls.
 * The walls in a grid squares are order as:
 *  ----> +X
 * |
 * |   +---2---+
 * V   |       |
 * +Z  3       1
 *     |       |
 *     +---0---+
 *
 * A wall is split into three sections along its height,
 * with each section defined by the heights of the grid points
 * at both ends of the wall:
 *
 *     h0[3] +----------+ h1[3]
 *           |          |         Section 2
 *     h0[2] +----------+ h1[2]
 *           |          |
 *           |          |         Section 1
 *           |          |
 * +Y  h0[1] +----------+ h1[1]
 * /\        |          |
 * |         |          |         Section 0
 * |   h0[0] +----------+ h1[0]
 *
 * Note that in general h0[0] does not need to be equal to h1[0],
 * h0[1] != h1[1], etc.
 *
 * Each section may be open or closed. If closed, the section is
 * filled with the wall's texture. If open, the section can be
 * looked through. Whether a section is open or not is given by
 * a single value for each wall. If section 0 is closed, then bit 0
 * of the value is set. If section 1 is closed, bit 1 is set, etc.
 *
 *   binary value      integer value     sections closed
 *                                          0   1   2
 *            000                  0        -   -   -
 *            001                  1        C   -   -
 *            010                  2        -   C   -
 *            011                  3        C   C   -
 *            100                  4        -   -   C
 *            101                  5        C   -   C
 *            110                  6        -   C   C
 *            111                  7        C   C   C
 *
 * If all sections of a wall are open, its texture may be
 * set to UNASSIGNED_UID.
 *
 * A square also has a floor and ceiling texture. If the
 * texture UID of a floor/ceiling is set to UNASSIGNED_UID,
 * there is no floor/ceiling.
 *
 * The floor is defined by the h[0] values of the four
 * grid points of the square. The ceiling is defined
 * by the h[3] values of the four grid points.
 *
 * Each wall, floor, and ceiling has a light value, which
 * indicates how brightly lit it is. It ranges from 0 (not
 * lit) to 255 (lit). How the light affects the colour of
 * the wall, floor, or ceiling is decided by the renderer.
 *
 * The UV coordinates at each vertex of walls, floors, and
 * ceilings are specified using a scale and offset applied
 * to the vertex position (x,y,z) in World coordinates.
 *  For walls 0 and 2:
 *    u = (x - offset[0]) * texscale
 *    v = (y - offset[1]) * texscale
 *  For walls 1 and 3:
 *    u = (z - offset[0]) * texscale
 *    v = (y - offset[1]) * texscale
 *  For floor and ceiling:
 *    u = (x - offset[0]) * texscale
 *    v = (z - offset[1]) * texscale
 */
// A World grid Point
struct WorldPoint
{
	float h[4];
};
// Texture on a wall, floor, or ceiling of a World Square
struct SquareTex
{
	UID texture;
	float scale;
	float offset[2];
};
// A World Square
struct WorldSquare
{
	unsigned char sectionsclosed[4];	// Which sections are closed, one value per wall
	SquareTex walltexture[4];			// Texture for each wall
	SquareTex floortexture;				// Floor texture, if any, of square
	SquareTex ceilingtexture;			// Ceiling texture, if any, of square
	unsigned char walllight[4];		// Amount of light on each wall
	unsigned char floorlight;			// Amount of light on floor, if any
	unsigned char ceilinglight;		// Amount of light on ceiling, if any
};

// A Gaggle message asking for a rectangular region of the World layout
// The asked for region is given by 2D indices into the grid, with
// x0 <= x1 and y0 <= y1. The World should reply with a WorldLayout
// message which includes all the squares with indices (x,y) with
// x0 <= x <= x1 and y0 <= y <= y1
struct AskWorldLayoutMessage : public GaggleMessage
{
	static void Fill(AskWorldLayoutMessage& msg, int x0, int y0, int x1, int y1);

	int box[2][2];
};

// A Gaggle message describing the World layout
struct WorldLayoutMessage : public GaggleMessage
{
	static WorldLayoutMessage* Create(unsigned int x0, unsigned int y0, unsigned int x1, unsigned int y1, WorldPoint*& worldpoints, WorldSquare*& worldsquares);
	static void Destroy(WorldLayoutMessage* msg);

	unsigned int box[2][2];			// The rectangular region of squares returned
	// Followed by (numsquares[0]+1) * (numsquares[1]+1) WorldPoint structures
	// Followed by numsquares[0] * numsquares[1] WorldSquare structures
	// where numsquares[0] = bbox[1][0] - bbox[0][0] + 1
	//       numsquares[1] = bbox[1][1] - bbox[0][1] + 1
	// The squares are stored by rows
};
#pragma pack(pop)


// Player actions
// Some actions have an associated value
typedef enum {
	ACTION_NONE = 0,					// No action, no value
	ACTION_KEY,							// A keyboard key as ASCII keycode, no value

	ACTION_MOVEFORWARD = 256,		// Basic movement, value is desired speed in metres/second
	ACTION_MOVEBACKWARD,
	ACTION_MOVELEFT,
	ACTION_MOVERIGHT,
	ACTION_TURNLEFT,					// Turning, value is desired rotation speed in radians/second
	ACTION_TURNRIGHT,
	ACTION_JUMP,						// Jumping, value is desired initial speed in metres/second
	ACTION_STOP,						// Stop moving

	ACTION_ACTIONBUTTON = 300,		// Action (fire) button, value is button number 0,1,2,3
} ActionCode;


// A Player
struct JoinPlayerMessage;
class Player : public Object
{
	protected:
		Player(const std::string& name);
	public:
		Player(const JoinPlayerMessage& msg, NetworkID networkid);
		virtual ~Player();
		NetworkID GetNetworkID() const;
		const std::string& GetName(void) const;

	protected:
		NetworkID networkid;
		std::string name;
};


// A HomeWorld
struct PlayerActionMessage;
struct TextMessage;
class HomeWorld : public Multiplayer<RakPeerInterface>, public World
{
	public:
		HomeWorld(const std::string& name, unsigned int numsquares_x, unsigned int numsquares_z, float squaresize, unsigned short maxconnects, unsigned short homeport);
		virtual ~HomeWorld();
		bool Join(const std::string& peerworldaddress, unsigned short peerport, unsigned long wait = 1000);
		void Leave(void);
		void ProcessPackets(void);

	protected:
		void SendToPlayers(const GaggleMessage& msg, NetworkID exclude = UNASSIGNED_PLAYER_ID);
		void SendToPlayer(const GaggleMessage& msg, const Player& player);

		// You probably want to override these
		virtual World* JoinPeerWorld(const WorldInfoMessage& msg, NetworkID peer_networkid);
		virtual void LeftPeerWorld(World* world);

		virtual Player* JoinPlayer(const JoinPlayerMessage& msg, NetworkID player_networkid);
		virtual void EnterPlayer(Player& player);
		virtual void PlayerHasLeft(Player* player);
		virtual void PlayerAction(Player& player, const PlayerActionMessage& msg);

		virtual void PlayerTexture(Player& player, const TextureMessage& msg);
		virtual void PlayerModel(Player& player, const ModelMessage& msg);

		virtual const Texture* AskTexture(Player& player, const AskTextureMessage& msg);
		virtual ObjectsMessage* AskObjects(Player& player, const AskObjectsMessage& msg);
		virtual const Model* AskModel(Player& player, const AskModelMessage& msg);

		virtual void PlayerTextMessage(Player& player, const TextMessage& msg);

		virtual WorldLayoutMessage* AskWorldLayout(Player& player, const AskWorldLayoutMessage& msg);

	private:
		void ReceiveConnectionRequestAccepted(Packet *packet, RakPeerInterface *interfaceType);
		void ReceiveNewIncomingConnection(Packet *packet, RakPeerInterface *interfaceType);
		void ReceiveDisconnectionNotification(Packet *packet, RakPeerInterface *interfaceType);
		void ReceiveConnectionLost(Packet *packet, RakPeerInterface *interfaceType);

		struct PlayerData;
		void CheckPlayerWelcome(PlayerData& playerdata);
		void ProcessUnhandledPacket(Packet *packet, unsigned char packetIdentifier, RakPeerInterface *interfaceType);

	private:
		RakPeerInterface *rakPeer;
		unsigned short maxconnects;
		unsigned short homeport;
		typedef std::map<NetworkID, World*> NetworkIDWorldMap;
		typedef std::pair<NetworkID, World*> NetworkIDWorldPair;
		NetworkIDWorldMap peers;
		struct PlayerData
		{
			Player* player;
			bool havetexture;
			bool havemodel;
			bool welcomed;
			bool isready;
		};
		typedef std::map<NetworkID, PlayerData> NetworkIDPlayerMap;
		typedef std::pair<NetworkID, PlayerData> NetworkIDPlayerPair;
		NetworkIDPlayerMap players;
};

#pragma pack(push)
#pragma pack(1)
// A Gaggle message containing some info aout a World
struct WorldInfoMessage : public GaggleMessage
{
	static void Fill(WorldInfoMessage& msg, const HomeWorld& world, unsigned char typeID, unsigned int size = sizeof(WorldInfoMessage));

	char name[64];						// Name of the world
};

// A Gaggle message from a HomeWorld introducting itself to a Player
struct WorldIntroMessage : public WorldInfoMessage
{
	static void Fill(WorldIntroMessage& msg, const HomeWorld& world, const Player& player);

	unsigned int numsquares[2];	// Number of squares in X and Z direction
	float squaresize;					// Size of a square in metres

	Vector initpos;					// Starting position of the Player
	float initheading;				// Starting heading of the Player
};

// A Gaggle message from a HomeWorld requesting to join a World
struct JoinWorldMessage : public WorldInfoMessage
{
	static void Fill(JoinWorldMessage& msg, const HomeWorld& world);
};

// A Gaggle message from a HomeWorld welcoming a peer World
struct WelcomeWorldMessage : public WorldInfoMessage
{
	static void Fill(WelcomeWorldMessage& msg, const World& world, const HomeWorld& homeworld);
};

// A Gaggle message from a HomeWorld welcoming a Player
struct WelcomePlayerMessage : public WorldInfoMessage
{
	static void Fill(WelcomePlayerMessage& msg, const Player& player, const HomeWorld& homeworld);

	UID uid;
};


// A Gaggle message from a World to a Player telling it to go
// to another World
struct ChangeWorldMessage : public GaggleMessage
{
	static void Fill(ChangeWorldMessage& msg, World& otherworld);

	NetworkID otherworld;
};
#pragma pack(pop)


// A HomePlayer
struct ScoreMessage;
class HomePlayer : public Multiplayer<RakClientInterface>, public Player
{
	public:
		HomePlayer(const std::string& name);
		virtual ~HomePlayer();

		bool Join(const std::string& worldaddress, unsigned short port, unsigned long wait = 3000);
		bool Join(NetworkID worldnetworkid);
		void Leave(void);
		void ProcessPackets(void);

	protected:
		void SendToWorld(const GaggleMessage& msg);

		// You probably want to override these
		virtual void JoinedWorld(const WelcomePlayerMessage& msg);
		virtual void LeftWorld(void);

		virtual void WorldObjects(const ObjectsMessage& msg);
		virtual void ObjectJoined(const ObjectJoinedMessage& msg);
		virtual void ObjectLeft(const ObjectLeftMessage& msg);
		virtual void ObjectState(const ObjectStateMessage& msg);

		virtual const Texture* AskTexture(void);
		virtual const Model* AskModel(void);

		virtual void WorldTexture(const TextureMessage& msg);
		virtual void WorldModel(const ModelMessage& msg);
		virtual void WorldLayout(const WorldLayoutMessage& msg);

		virtual void WorldTextMessage(const TextMessage& msg);
		virtual void WorldScoreMessage(const ScoreMessage& msg);

		virtual void WorldIntro(const WorldIntroMessage& msg);

		virtual void ChangeWorld(const ChangeWorldMessage& msg);

	private:
		void ReceiveConnectionRequestAccepted(Packet *packet, RakClientInterface *interfaceType);
		void ReceiveDisconnectionNotification(Packet *packet, RakClientInterface *interfaceType);

		void ProcessUnhandledPacket(Packet *packet, unsigned char packetIdentifier, RakClientInterface *interfaceType);

	private:
		RakClientInterface *rakClient;
};

#pragma pack(push)
#pragma pack(1)
// A Gaggle message from a Player requesting to join a World
struct JoinPlayerMessage : public IdentifiedMessage
{
	static void Fill(JoinPlayerMessage& msg, const HomePlayer& player);

	char name[64];
};

// A Gaggle message from a Player to a World requesting to do an Action
struct PlayerActionMessage : public IdentifiedMessage
{
	static void Fill(PlayerActionMessage& msg, const Player& player, ActionCode code, float value);

	ActionCode code;
	float value;
};

// A Gaggle message from a Player to a World, or a World to a Player
// giving some text message
struct TextMessage : public GaggleMessage
{
	static void Fill(TextMessage& msg, const std::string& text);

	char text[256];
};

// A Gaggle message from a Player to a World saying the player is
// now ready to play
struct PlayerReadyMessage : public GaggleMessage
{
	static void Fill(PlayerReadyMessage& msg);
};

// A Gaggle message from a World to a Player giving some sort of
// scoring text
struct ScoreMessage : public GaggleMessage
{
	static void Fill(ScoreMessage& msg, const std::string& text);

	char text[64];
};

#pragma pack(pop)

}
