//
// Gaggle objects
// by Jarno van der Linden  jvan006@cs.auckland.ac.nz
//

#include "gaggle.h"

#include <cmath>

using namespace Gaggle;
using namespace std;

Object::Object()
	: Identified()
{
	starttime = RakNetGetTime();
	endtime = starttime;
	pos.x = pos.y = pos.z = 0;
	dpos.x = dpos.y = dpos.z = 0;
	ddpos.x = ddpos.y = ddpos.z = 0;
	heading = 0;
	dheading = 0;

	startframe = endframe = 0;
	animfps = 0;
	animstarttime = starttime;

	textureuid = UNASSIGNED_UID;
	modeluid = UNASSIGNED_UID;

	camfixedpos = camfixedrot = false;
}

Object::Object(const ObjectJoinedMessage& msg)
	: Identified(msg)
{
	starttime = RakNetGetTime();
	endtime = starttime;
	pos.x = pos.y = pos.z = 0;
	dpos.x = dpos.y = dpos.z = 0;
	ddpos.x = ddpos.y = ddpos.z = 0;
	heading = 0;
	dheading = 0;

	startframe = endframe = 0;
	animfps = 0;
	animstarttime = starttime;

	textureuid = msg.textureuid;
	modeluid = msg.modeluid;

	camfixedpos = (msg.camfixed & 1) ? true : false;
	camfixedrot = (msg.camfixed & 2) ? true : false;
}

Object::Object(const IdentifiedMessage& msg)
	: Identified(msg)
{
	starttime = RakNetGetTime();
	endtime = starttime;
	pos.x = pos.y = pos.z = 0;
	dpos.x = dpos.y = dpos.z = 0;
	ddpos.x = ddpos.y = ddpos.z = 0;
	heading = 0;
	dheading = 0;

	startframe = endframe = 0;
	animfps = 0;
	animstarttime = starttime;

	textureuid = UNASSIGNED_UID;
	modeluid = UNASSIGNED_UID;

	camfixedpos = camfixedrot = false;
}

Object::~Object()
{
}


// Get the start time of the last state update
unsigned long Object::GetStartTime(void) const
{
	return starttime;
}

// Get the end time of the last state update
unsigned long Object::GetEndTime(void) const
{
	return endtime;
}

// Get the position of the last state update
void Object::GetPos(float& x, float& y, float& z) const
{
	x = pos.x;
	y = pos.y;
	z = pos.z;
}
void Object::GetPos(Vector& p) const
{
	GetPos(p.x, p.y, p.z);
}

// Get the velocity vector of the last state update
void Object::GetDPos(float& dx, float& dy, float& dz) const
{
	dx = dpos.x;
	dy = dpos.y;
	dz = dpos.z;
}
void Object::GetDPos(Vector& dp) const
{
	GetDPos(dp.x, dp.y, dp.z);
}

// Get the acceleration vector of the last state update
void Object::GetDDPos(float& ddx, float& ddy, float& ddz) const
{
	ddx = ddpos.x;
	ddy = ddpos.y;
	ddz = ddpos.z;
}
void Object::GetDDPos(Vector& ddp) const
{
	GetDDPos(ddp.x, ddp.y, ddp.z);
}

// Compute a current position based on the last state
void Object::GetExtrapolatedPos(unsigned long curtime, float& x, float& y, float& z) const
{
	// p' = p  + dp * dt + 1/2 * ddp * dt^2
	// Clip current time to endtime
	if(curtime > endtime)
		curtime = endtime;
	float dtime;
	if(curtime >= starttime)
		dtime = (curtime - starttime) / 1000.f;
	else
		dtime = -((starttime - curtime) / 1000.f);
	x = pos.x + dpos.x * dtime + 0.5f * ddpos.x * dtime*dtime;
	y = pos.y + dpos.y * dtime + 0.5f * ddpos.y * dtime*dtime;
	z = pos.z + dpos.z * dtime + 0.5f * ddpos.z * dtime*dtime;
}

void Object::GetExtrapolatedPos(unsigned long curtime, Vector& p) const
{
	GetExtrapolatedPos(curtime, p.x, p.y, p.z);
}

// Compute a current change in position based on the last state
void Object::GetExtrapolatedDPos(unsigned long curtime, float& dx, float& dy, float& dz) const
{
	// dp' = dp + ddp * dt
	// Clip current time to endtime
	if(curtime > endtime)
		curtime = endtime;
	float dtime;
	if(curtime >= starttime)
		dtime = (curtime - starttime) / 1000.f;
	else
		dtime = -((starttime - curtime) / 1000.f);
	dx = dpos.x + ddpos.x * dtime;
	dy = dpos.y + ddpos.y * dtime;
	dz = dpos.z + ddpos.z * dtime;
}

void Object::GetExtrapolatedDPos(unsigned long curtime, Vector& dp) const
{
	GetExtrapolatedDPos(curtime, dp.x, dp.y, dp.z);
}

// Get the heading of the last state update
float Object::GetHeading(void) const
{
	return heading;
}

// Get the heading rate of change the last state update
float Object::GetDHeading(void) const
{
	return dheading;
}

// Compute a current heading based on the last state
float Object::GetExtrapolatedHeading(unsigned long curtime) const
{
	// h' = h + dh * dt
	if((endtime != ULONG_MAX) && (curtime > endtime))
		curtime = endtime;
	float dtime = (curtime - starttime) / 1000.f;
	return heading + dheading * dtime;
}

// Get the start and end frame of the current animation loop
void Object::GetAnimLoop(unsigned int& startframe_out, unsigned int& endframe_out) const
{
	startframe_out = startframe;
	endframe_out = endframe;
}

// Get the animation speed
float Object::GetAnimFPS(void) const
{
	return animfps;
}

// Get the start time of the animation loop
unsigned long Object::GetAnimStartTime(void) const
{
	return animstarttime;
}

// Get the current frame given the current time,
// with fractional part of interpolation.
// The returned value is in the range [startframe, endframe+1)
// At the end, interpolation should be from the endframe state
// to the startframe state.
float Object::GetExtrapolatedFrame(unsigned long curtime)
{
	if(curtime < animstarttime)
		return (float)startframe;

	float dtime = (curtime - animstarttime) / 1000.f;
	return (float)startframe + fmodf(dtime * animfps, (float)(endframe - startframe + 1));
}


void Object::SetStartTime(unsigned long starttime_in)
{
	starttime = starttime_in;
}

void Object::SetEndTime(unsigned long endtime_in)
{
	endtime = endtime_in;
}

void Object::SetPos(float x, float y, float z)
{
	pos.x = x;
	pos.y = y;
	pos.z = z;
}

void Object::SetPos(const Vector& p)
{
	SetPos(p.x, p.y, p.z);
}

void Object::SetDPos(float dx, float dy, float dz)
{
	dpos.x = dx;
	dpos.y = dy;
	dpos.z = dz;
}

void Object::SetDPos(const Vector& dp)
{
	SetDPos(dp.x, dp.y, dp.z);
}

void Object::SetDDPos(float ddx, float ddy, float ddz)
{
	ddpos.x = ddx;
	ddpos.y = ddy;
	ddpos.z = ddz;
}

void Object::SetDDPos(const Vector& ddp)
{
	SetDDPos(ddp.x, ddp.y, ddp.z);
}

void Object::SetHeading(float heading_in)
{
	heading = heading_in;
}

void Object::SetDHeading(float dheading_in)
{
	dheading = dheading_in;
}

// Set the start and end frame of the current animation loop
void Object::SetAnimLoop(unsigned int startframe_in, unsigned int endframe_in)
{
	startframe = startframe_in;
	endframe = endframe_in;
}

// Set the animation speed
void Object::SetAnimFPS(float animfps_in)
{
	animfps = animfps_in;
}

// Set the start time of the animation loop
void Object::SetAnimStartTime(unsigned long animstarttime_in)
{
	animstarttime = animstarttime_in;
}

// Set a complete state
void Object::SetState(const ObjectStateMessage& msg)
{
	SetStartTime(msg.starttime);
	SetEndTime(msg.endtime);
	SetPos(msg.pos);
	SetDPos(msg.dpos);
	SetDDPos(msg.ddpos);
	SetHeading(msg.heading);
	SetDHeading(msg.dheading);
	SetAnimLoop(msg.startframe, msg.endframe);
	SetAnimFPS(msg.animfps);
	SetAnimStartTime(msg.animstarttime);
}


// Object Model and Texture
UID Object::GetModelUID(void) const
{
	return modeluid;
}

UID Object::GetTextureUID(void) const
{
	return textureuid;
}


void Object::SetModelUID(UID modeluid_in)
{
	modeluid = modeluid_in;
}


void Object::SetTextureUID(UID textureuid_in)
{
	textureuid = textureuid_in;
}


// Objects can have their position or rotation fixed relative
// to the camera.
void Object::GetCamFixed(bool& camfixedpos_out, bool& camfixedrot_out) const
{
	camfixedpos_out = camfixedpos;
	camfixedrot_out = camfixedrot;
}

void Object::SetCamFixed(bool camfixedpos_in, bool camfixedrot_in)
{
	camfixedpos = camfixedpos_in;
	camfixedrot = camfixedrot_in;
}


// A message from a World to all Players informing about an Object state
void ObjectStateMessage::Fill(ObjectStateMessage& msg, const Object& object)
{
	IdentifiedMessage::Fill(msg, object, ID_OBJECTSTATE_MESSAGE, sizeof(ObjectStateMessage));

	msg.worldtime = msg.timestamp;
	msg.starttime = object.GetStartTime();
	msg.endtime = object.GetEndTime();
	object.GetPos(msg.pos);
	object.GetDPos(msg.dpos);
	object.GetDDPos(msg.ddpos);
	msg.heading = object.GetHeading();
	msg.dheading = object.GetDHeading();
	object.GetAnimLoop(msg.startframe, msg.endframe);
	msg.animfps = object.GetAnimFPS();
	msg.animstarttime = object.GetAnimStartTime();
}

// A message from a World to every player in the World telling
// them about a new object having joined
void ObjectJoinedMessage::Fill(ObjectJoinedMessage& msg, const Object& object)
{
	IdentifiedMessage::Fill(msg, object, ID_OBJECTJOINED_MESSAGE, sizeof(ObjectJoinedMessage));
	msg.textureuid = object.GetTextureUID();
	msg.modeluid = object.GetModelUID();
	bool camfixedpos, camfixedrot;
	object.GetCamFixed(camfixedpos, camfixedrot);
	msg.camfixed = 0;
	if(camfixedpos) msg.camfixed |= 1;
	if(camfixedrot) msg.camfixed |= 2;
}

// A message from a World to every player in the World telling
// them about an object having left
void ObjectLeftMessage::Fill(ObjectLeftMessage& msg, const Object& object)
{
	IdentifiedMessage::Fill(msg, object, ID_OBJECTLEFT_MESSAGE, sizeof(ObjectLeftMessage));
}

// A message asking for all objects currently in the World
void AskObjectsMessage::Fill(AskObjectsMessage& msg)
{
	GaggleMessage::Fill(msg, ID_ASKOBJECTS_MESSAGE, sizeof(AskObjectsMessage));
}

// A message telling a Player about all the objects currently in the World
ObjectsMessage* ObjectsMessage::Create(unsigned int numobjects, ObjectData*& objectdata)
{
	unsigned int objectsmsgsize = sizeof(ObjectsMessage) + numobjects * sizeof(ObjectData);
	ObjectsMessage* objectsmsg = (ObjectsMessage *)new unsigned char[objectsmsgsize];
	GaggleMessage::Fill(*objectsmsg, ID_OBJECTS_MESSAGE, objectsmsgsize);

	objectsmsg->numobjects = numobjects;
	objectdata = (ObjectData *)(((unsigned char *)objectsmsg) + sizeof(ObjectsMessage));

	return objectsmsg;
}

void ObjectsMessage::Destroy(ObjectsMessage* msg)
{
	delete[] msg;
}
